print("import: 'plotly_resampler'")
import plotly_resampler

print("import: 'plotly_resampler.figure_resampler'")
import plotly_resampler.figure_resampler

print("import: 'plotly_resampler.aggregation.algorithms.lttbc'")
import plotly_resampler.aggregation.algorithms.lttbc

